﻿#LoiLoNote School Sideloading Setup Version 2.1
param(
    [ValidateSet('Select', 'Install','Uninstall','ShowLog','ProvisionedInstall','ProvisionedUninstall', 'ElevatedProvisionedInstall', 'ElevatedProvisionedUninstall', 'RemoteInstall')]
    [string] $Action = 'Select',
    [string] $Packages,
    [switch] $Always,
    [switch] $Force,
    [switch] $EnsureInstall
)

$ErrorActionPreference = "Stop"

Set-PSDebug -Strict

$LoiLoPackageNameBase = "LoiLoinc.51956F723A3DD.s"
$DefaultPackageNameQuery = $LoiLoPackageNameBase + "*"
$CompanyName = "LoiLo"
$ScriptPath = (Get-Variable MyInvocation).Value.MyCommand.Path
$ScriptDirPath = Split-Path $ScriptPath -Parent
$ScriptName = $(Get-ChildItem $ScriptPath).BaseName
$WorkDirPath = Join-Path "$env:LOCALAPPDATA" (Join-Path $CompanyName $ScriptName)
$TempDirPath = Join-Path $WorkDirPath "Temp"
$LogDirPath = Join-Path $WorkDirPath "Logs"
$SettingsFileCopyTries = 4

#region Log

$LogMaxLines = 1000
$LogFileName = "log.txt"
$LogFilePath = Join-Path $LogDirPath $LogFileName
$LogBackupFileName = "log.txt.back"
$LogBackupFilePath = Join-Path $LogDirPath $LogBackupFileName

function Append-Log{

    param(
        [ValidateSet("INFO", "ERROR", ignorecase=$true)]
        [string] $type,
        [string] $msg
    )

    $now = Get-Date -Format "yyyy/MM/dd HH:mm:ss"
    $message = ("$now {0,-5} $msg" -f $type)

    if ($type -ne "ERROR") {
        Write-Host $message
    } else {
        Write-Host $message -ForegroundColor Red
    }

    if(-not (Test-Path $LogDirPath -PathType Container)){
        New-Item $LogDirPath -ItemType Directory -Force | Out-Null
    }

    if(Test-Path $LogFilePath -PathType Leaf){
        $line = (Get-Content $LogFilePath | Measure-Object –Line).Lines
        if( $line -ge $LogMaxLines ){
            if(Test-Path $LogBackupFilePath -PathType Leaf){
                Remove-Item $LogBackupFilePath -Force 
            }        
            Rename-Item $LogFilePath -NewName $LogBackupFileName 
        }
    }

    Write-Output $message | Out-File ${LogFilePath} -Append -Encoding utf8
}

#endregion

#region Zip

function Expand-Zip {
    param([string] $zipFilePath, [string] $dstDirPath)

    New-Item $dstDirPath -ItemType Directory -Force | Out-NUll

    # Zipファイルにファイルおよびフォルダを投入
    $shell = New-Object -com Shell.Application
    $zip = $shell.NameSpace($zipFilePath)
    foreach($item in $zip.items())
    {
        $shell.Namespace($dstDirPath).copyhere($item)
    }
}

#endregion 

#region Logic

function Find-File {
    param( [string] $dirPath, [string] $filePattern)

    $queryPath = (Join-Path $dirPath "*.*")
    $foundItems = @(Get-ChildItem $queryPath -Include $filePattern | Where-Object { $_.Mode -NotMatch "d" })
    if($foundItems.Length -le 0) {
        return $null
    }
    return $foundItems[0]
}

function Get-TargetPackageInfo {
    param([string] $dirName)
    $packageJsonFilePath = Join-Path $dirName "package.json"
    $packageJsonContent = Get-Content -Path $packageJsonFilePath -Encoding UTF8 -Raw 
    $info = ConvertFrom-Json $packageJsonContent
    return $info
}

function Compare-Version {
    param([string] $version1, [string] $version2)

    $array1 = @($version1 -split "." , 0 , "simplematch")
    $array2 = @($version2 -split "." , 0 , "simplematch")

    for($i = 0; $i -lt 4; $i++) {
        $contains1 = ($array1.Count - 1) -ge $i;
        $contains2 = ($array2.Count - 1) -ge $i;
        if((-not $contains1) -and (-not $contains2)) {
            return 0;
        } elseif(-not $contains1) {
            return -1;
        } elseif(-not $contains2) {
            return 1;
        }

        $num1 = [int]$array1[$i]
        $num2 = [int]$array2[$i]

        if($num1 -gt $num2) {
            return 1;
        } elseif($num1 -lt $num2) {
            return -1;
        }
    }

    return 0;
}

function Compare-Generation {
    param([string]$leftPackageName, [string]$rightPackageName)
    $leftNum = [int16]::Parse($leftPackageName.Replace($LoiLoPackageNameBase, ""))
    $rightNum = [int16]::Parse($rightPackageName.Replace($LoiLoPackageNameBase, ""))

    if ($leftNum -gt $rightNum) {
        return 1
    }
    elseif ($leftNum -lt $rightNum) {
        return -1
    }
    elseif ($leftNum -eq $rightNum) {
        return 0
    }
    else {
        throw "Cannot compare $leftPackageName with $rightPackageName"
    }
}

function Get-DependencyPackages {
    param([string] $dirPath)

    $dependenciesDirPath = (Join-Path $dirPath "Dependencies")
    $dependencyPackages = @()
    if (Test-Path $dependenciesDirPath) {
        $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "*.appx") | Where-Object { $_.Mode -NotMatch "d" }

        if (($Env:Processor_Architecture -eq "x86" -or $Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $dependenciesDirPath "x86"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "x86\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        } 
        if (($Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $dependenciesDirPath "x64"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "x64\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        }
        if (($Env:Processor_Architecture -eq "arm") -and (Test-Path (Join-Path $dependenciesDirPath "arm"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "arm\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        }
    }

    return $dependencyPackages
}

function Try-Copy {
    param([string] $from, [string] $to)

    try {
        Copy-Item $from $to -Force
        return $true
    }
    catch {
        Append-Log ERROR $_.Exception.Message
        return $false
    }
}

function Attempt-Migration {
    param(
        [string] $from,
        [string] $to,
        [int] $tries
    )

    $copyTries = $tries
    while ( $copyTries -gt 0 ) {
        Start-Sleep -Milliseconds 500
        if ( Try-Copy -from $from -to $to ) {
            Append-Log INFO "Copied previous settings file from `"$from`" to `"$to`""
            break
        }

        if ( $copyTries -eq 1 ) {
            Append-Log ERROR "Could not copy previous settings file from `"$from`" to `"$to`""
            break
        }

        Append-Log INFO "Could not copy previous settings file. Will try again."
        $copyTries--
    }
}

function Install-Package {
    param([string] $dirPath)

    $bundleFilePath = Find-File $dirPath "*.appxbundle"
    if($bundleFilePath -eq $null) {
        throw "No bundle file in package."
    }

    $dependencyPackages = Get-DependencyPackages $dirPath

    try {
        if ($dependencyPackages.FullName.Count -gt 0) {
            Append-Log INFO "Found dependency($dependencyPackages)"
            Add-AppxPackage "$bundleFilePath" -DependencyPath $dependencyPackages.FullName -ForceApplicationShutdown
        } else {
            Append-Log INFO "No dependencies found."
            Add-AppxPackage "$bundleFilePath" -ForceApplicationShutdown
        }
    }
    catch {
        Append-Log ERROR "Package installation resulted in an error."
        throw
    }
}

function Install-ProvisionedPackage {
    param([string] $dirPath)
    
    $bundleFilePath = Find-File $dirPath "*.appxbundle"
    if($bundleFilePath -eq $null) {
        throw "No bundle file in package."
    }

    $dependencyPackages = Get-DependencyPackages $dirPath

    if ($dependencyPackages.FullName.Count -gt 0) {
        Append-Log INFO "Found dependency($dependencyPackages)"
        $image = Add-AppxProvisionedPackage -Online -PackagePath "$bundleFilePath" -DependencyPackagePath $dependencyPackages.FullName -SkipLicense
        Append-Log INFO ($image | Format-List | Out-String)
    } else {
        $image = Add-AppxProvisionedPackage -Online -PackagePath "$bundleFilePath" -SkipLicense
        Append-Log INFO ($image | Format-List | Out-String)
    }
}

function Uninstall-Package {
    param([string] $packageName)
    $package = Get-AppxPackage -Name $packageName
    if($package) {
        Append-Log INFO "Uninstalling $packageName (v.$($package.Version))..."
        Remove-AppxPackage -Package $package.PackageFullName
        Append-Log INFO "Uninstall complete!"
    } else {
        Append-Log INFO "Could not find an installed package named `"$packageName`""
    }
}

function Uninstall-ProvisionedPackage{
    param([string] $packageNameQuery)
    $packages = @(Get-AppxProvisionedPackage -Online | Where-Object DisplayName -like $packageNameQuery)

    if($packages) {
        foreach( $package in $packages ) {
            $packageFullname = $package.PackageName
            Append-Log INFO "Uninstalling ProvisionedPackage($packageFullname)..."
            $image = Remove-AppxProvisionedPackage -Online -PackageName $packageFullname
            Append-Log INFO ($image | Format-List | Out-String)
        }
        Append-Log INFO "Provisioned Package Uninstall Complete!" 
    } else {
        Append-Log INFO "Could not find an installed ProvisionedPackage named `"$packageName`""
    }   
}

function Run-RemoteInstall {

    if($Env:Processor_Architecture -eq "arm") {
        throw "RemoteInstall not supported with ARM processors."
    }

    $remoteTempDirPath = Join-Path $TempDirPath "Remote"
    New-Item $remoteTempDirPath -ItemType Directory -Force | Out-NUll
    $remotePackageInfoFilePath = Join-Path $remoteTempDirPath "package.json"

    Append-Log INFO "Checking remote version..."

    $remotePackageInfoUri = "http://assets.loilo.tv/loilonote/windows/sideloading/package.json"
    Invoke-RestMethod $remotePackageInfoUri -OutFile $remotePackageInfoFilePath
    $packageJsonContent = Get-Content -Path $remotePackageInfoFilePath -Encoding UTF8 -Raw 
    $targetPackageInfo = ConvertFrom-Json $packageJsonContent
    $packageName = $targetPackageInfo.Name
    $targetVersion = $targetPackageInfo.Version
    $targetPublisherId = $targetPackageInfo.PublisherId
    $targetPackageFamilyName = "$($packageName)_$($targetPublisherId)"

    Append-Log INFO "Remote package: $packageName (v.$targetVersion)"

    $packageSortProperty = @{Expression={[int16]::Parse($_.Name.Replace($LoiLoPackageNameBase, ""))}; Descending=$true}
    $installedPackages = Get-AppxPackage -Name $DefaultPackageNameQuery | Sort-Object -Property $packageSortProperty

    if ( $installedPackages.Count -gt 0 ) {
        foreach( $installedPackage in $installedPackages ) {

            $genCompareResult = Compare-Generation $installedPackage.Name $packageName

            if( $genCompareResult -gt 0 ) {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of a newer generation."

            } elseif( $genCompareResult -eq 0 ) {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of the same generation as the target package."
                $verCompareResult = Compare-Version $installedPackage.Version $targetVersion
                
                if( $verCompareResult -gt 0 ) {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is of a newer version than target package."

                    if ( $Always ) {
                        Append-Log INFO "The `"-Always`" flag has been deprecated. To force a downgrade, uninstall $($installedPackage.Name), then install $packageName."
                    } 

                    Append-Log INFO "Target package $packageName will not be installed."
                    return

                } elseif( $verCompareResult -eq 0 ) {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is the same version as target package. No installation required."
                    return

                } else {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is an older version than target package."
                    $installedPackagePublisherId = $installedPackage.PublisherId

                    if($installedPackagePublisherId -ne $targetPublisherId) {
                        Append-Log INFO "Currently installed $($installedPackage.Name) has a different publisher hash."
                    }
                }
            } else {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of an older generation."
            }
        }
    } else {
        Append-Log INFO "Previously installed sideloading version of LoiLoNote School not found."
    }

    Append-Log INFO "Downloading version=$targetVersion ..."
    $remotePackageFileName = "LoiLoNoteSetup_$targetVersion.zip"
    $remotePackageFilePath = Join-Path $remoteTempDirPath $remotePackageFileName

    $remotePackageUri = "http://assets.loilo.tv/loilonote/windows/sideloading/$remotePackageFileName"
    Invoke-RestMethod $remotePackageUri -OutFile $remotePackageFilePath

    $remoteUnzipDirPath = Join-Path $remoteTempDirPath "LoiLoNoteSetup"
    New-Item $remoteUnzipDirPath -ItemType Directory -Force | Out-NUll
    Expand-Zip $remotePackageFilePath $remoteUnzipDirPath

    $packageDirPath = Join-Path $remoteUnzipDirPath "Packages"

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if( -not( Test-Path $indexFilePath ) ) {
        throw "Invalid path: $indexFilePath"
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if( $indexFileContent.Length -le 0 ) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if( -not( Test-Path $targetDirPath ) ) {
        throw "Invalid path: $targetDirPath"
    }

    $localPackagesDir = Join-Path $env:LOCALAPPDATA "Packages"

    $settingsMigrationRequired = $false
    $migrationSourceSettingsFilePath = ""
    if ( $installedPackages.Count -gt 0 ) {
        $latestInstalledPackageFamilyName = $installedPackages[0].PackageFamilyName
        if ( $latestInstalledPackageFamilyName -ne $targetPackageFamilyName ) {

            $settingsMigrationRequired = $true
            $migrationSourcePackageDataDir = Join-Path $localPackagesDir $latestInstalledPackageFamilyName
            $migrationSourceSettingsFilePath = Join-Path $migrationSourcePackageDataDir "Settings\settings.dat"
        }
    }

    Append-Log INFO "Installing $($packageName) (v.$targetVersion)"
    Install-Package $targetDirPath

    if ( $settingsMigrationRequired ) {
        
        $targetPackageDataDir = Join-Path $localPackagesDir $targetPackageFamilyName
        $targetSettingsFilePath = Join-Path $targetPackageDataDir "Settings\settings.dat"

        if( ( Test-Path $migrationSourceSettingsFilePath ) -and ( Test-Path $targetSettingsFilePath ) ) {
            Append-Log INFO "Copying previous settings file to newly installed package..."
            Attempt-Migration -from $migrationSourceSettingsFilePath -to $targetSettingsFilePath -tries $SettingsFileCopyTries
        }
        else {
            Append-Log ERROR "Unable to copy previous settings file from `"$migrationSourceSettingsFilePath`" to `"$targetSettingsFilePath`""
        }
    }

    Append-Log INFO "Remote Install Complete!"
}

function Run-Install {
    param([string] $packageDirPath)

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if( -not( Test-Path $indexFilePath )) {
        throw "Invalid path: $indexFilePath"
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if( $indexFileContent.Length -le 0 ) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if( -not( Test-Path $targetDirPath ) ) {
        throw "Invalid path: $targetDirPath"
    }

    $targetPackageInfo = Get-TargetPackageInfo $targetDirPath
    $packageName = $targetPackageInfo.Name
    $targetVersion = $targetPackageInfo.Version
    $targetPublisherId = $targetPackageInfo.PublisherId
    $targetPackageFamilyName = "$($packageName)_$($targetPublisherId)"

    $packageSortProperty = @{Expression={[int16]::Parse($_.Name.Replace($LoiLoPackageNameBase, ""))}; Descending=$true}
    $installedPackages = Get-AppxPackage -Name $DefaultPackageNameQuery | Sort-Object -Property $packageSortProperty

    if ( $installedPackages.Count -gt 0 ) {
        foreach( $installedPackage in $installedPackages ) {

            $genCompareResult = Compare-Generation $installedPackage.Name $packageName

            if( $genCompareResult -gt 0 ) {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of a newer generation."

            } elseif($genCompareResult -eq 0) {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of the same generation as the target package."
                $verCompareResult = Compare-Version $installedPackage.Version $targetVersion
                
                if($verCompareResult -gt 0) {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is of a newer version than target package."

                    if ( $Always ) {
                        Append-Log INFO "The `"-Always`" flag has been deprecated. To force a downgrade, uninstall $($installedPackage.Name), then install $packageName."
                    }

                    Append-Log INFO "Target package $packageName will not be installed."
                    return

                } elseif($verCompareResult -eq 0) {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is the same version as target package. No installation required."
                    return

                } else {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is an older version than target package."
                    $installedPackagePublisherId = $installedPackage.PublisherId

                    if($installedPackagePublisherId -ne $targetPublisherId) {
                        Append-Log INFO "Currently installed $($installedPackage.Name) has a different publisher hash."
                    }
                }
            } else {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of an older generation."
            }
        }
    } else {
        Append-Log INFO "Previously installed sideloading version of LoiLoNote School not found."
    }

    $copyDirPath = Join-Path $TempDirPath $targetDirName
    Copy-Item -Path $targetDirPath -Destination $copyDirPath -Force -Recurse | Out-Null

    $localPackagesDir = Join-Path $env:LOCALAPPDATA "Packages"

    $settingsMigrationRequired = $false
    $migrationSourceSettingsFilePath = ""
    if ( $installedPackages.Count -gt 0 ) {
        $latestInstalledPackageFamilyName = $installedPackages[0].PackageFamilyName
        if ( $latestInstalledPackageFamilyName -ne $targetPackageFamilyName ) {

            $settingsMigrationRequired = $true
            $migrationSourcePackageDataDir = Join-Path $localPackagesDir $latestInstalledPackageFamilyName
            $migrationSourceSettingsFilePath = Join-Path $migrationSourcePackageDataDir "Settings\settings.dat"
        }
    }

    Append-Log INFO "Installing $($packageName) (v.$targetVersion)"
    Install-Package $copyDirPath

    if( $settingsMigrationRequired ) {

        $targetPackageDataDir = Join-Path $localPackagesDir $targetPackageFamilyName
        $targetSettingsFilePath = Join-Path $targetPackageDataDir "Settings\settings.dat"

        if ( ( Test-Path $migrationSourceSettingsFilePath ) -and ( Test-Path $targetSettingsFilePath) ){
            Append-Log INFO "Copying previous settings file to newly installed package..."
            Attempt-Migration -from $migrationSourceSettingsFilePath -to $targetSettingsFilePath -tries $SettingsFileCopyTries
        }
        else {
            Append-Log ERROR "Unable to copy previous settings file from `"$migrationSourceSettingsFilePath`" to `"$targetSettingsFilePath`""
        }
    }

    Append-Log INFO "Installation Complete!"
}


function Run-ProvisionedInstall {
    param([string] $packageDirPath)

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if(-not(Test-Path $indexFilePath)) {
        throw "Invalid path: $indexFilePath"
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if($indexFileContent.Length -le 0) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if(-not(Test-Path $targetDirPath)) {
		throw "Invalid path: $targetDirPath"
    }

    $targetPackageInfo = Get-TargetPackageInfo $targetDirPath
    $packageName = $targetPackageInfo.Name
    Uninstall-ProvisionedPackage $DefaultPackageNameQuery


    $copyDirPath = Join-Path $TempDirPath $targetDirName
    Copy-Item -Path $targetDirPath -Destination $copyDirPath -Force -Recurse | Out-Null

    Append-Log INFO "Installing ProvisionedPackage..."
    Install-ProvisionedPackage $copyDirPath
    Append-Log INFO "Provisioned Install Complete!"
}

function Ensure-TempDir {
    Append-Log INFO "Preparing $TempDirPath"
    if(Test-Path $TempDirPath)
    {
        Remove-Item $TempDirPath -Recurse -Force
    }
    New-Item $TempDirPath -ItemType Directory -Force | Out-NUll
}

function Run{

    $running = $Action;
    if($running -eq 'Select') {
        $input = [int](Read-Host "Please select option below.`n0:Exit, 1:Install, 2:Uninstall, 3:ShowLog, 4:ProvisionedInstall, 5:ProvisionedUninstall, 6:RemoteInstall`n=>")
        if($input -eq 1) {
            $running = 'Install';
        } elseif ($input -eq 2) {
            $running = 'Uninstall';
        } elseif ($input -eq 3) {
            $running = 'ShowLog';
        } elseif ($input -eq 4) {
            $running = 'ProvisionedInstall';
        } elseif ($input -eq 5) {
            $running = 'ProvisionedUninstall';
        } elseif ($input -eq 6) {
            $running = 'RemoteInstall';
        } else {
            Append-Log INFO "Exit."
            return $true
        }
    }

    Append-Log INFO "Action=$running"

    if($running -eq 'Install') {
        Ensure-TempDir
        $packageDirPath = Join-Path $ScriptDirPath "Packages"
        if( $Packages -ne "" ){
            $packageDirPath = $Packages
        }
        Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath )"
        Run-Install $packageDirPath

        return $false
    } elseif ($running -eq 'Uninstall') {
        
		$uninstallablePackages = Get-AppxPackage -Name $DefaultPackageNameQuery

		if($uninstallablePackages.Count -gt 0){
            foreach ($package in $uninstallablePackages) {
                Uninstall-Package $package.Name
            }
			
		}
		else {
			Append-Log INFO "Uninstallable package not found."
		}
		
        return $false

    } elseif ($running -eq 'ShowLog') {
        Start-Process "$LogDirPath" | Out-Null

        return $true
    } elseif ($running -eq 'ProvisionedInstall') {

        $relaunchArgs = '-ExecutionPolicy Unrestricted -file "' + $ScriptPath + '" -Action ElevatedProvisionedInstall'

        if($Force) {
            $relaunchArgs += ' -Force'
        }

        if( $Packages -ne "" ) {
            $relaunchArgs += ' -Packages "' + $Packages + '"'
        }
        
        Start-Process "$PsHome\PowerShell.exe" -Verb RunAs -ArgumentList $relaunchArgs -PassThru | Out-Null

        return $true
    } elseif($running -eq 'ProvisionedUninstall') {
        $relaunchArgs = '-ExecutionPolicy Unrestricted -file "' + $ScriptPath + '" -Action ElevatedProvisionedUninstall'
        Start-Process "$PsHome\PowerShell.exe" -Verb RunAs -ArgumentList $relaunchArgs -PassThru | Out-Null

        return $true
    } elseif($running -eq "ElevatedProvisionedInstall") {
        Ensure-TempDir
        $packageDirPath = Join-Path $ScriptDirPath "Packages"
        if( $Packages -ne "" ){
            $packageDirPath = $Packages
        }
        Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath )"
        Run-ProvisionedInstall $packageDirPath

        return $false
    } elseif ($running -eq "ElevatedProvisionedUninstall"){

		Uninstall-ProvisionedPackage $DefaultPackageNameQuery

        return $false
    } elseif ($running -eq "RemoteInstall"){
        Ensure-TempDir
        Append-Log INFO "Param( Always=$Always )"

        try {            
            Run-RemoteInstall
        } catch {
            Append-Log ERROR $_.Exception.Message

            if($EnsureInstall){
                
                Append-Log INFO "Installing local package..."

                Ensure-TempDir
                $packageDirPath = Join-Path $ScriptDirPath "Packages"
                if( $Packages -ne "" ){
                    $packageDirPath = $Packages
                }

                $Always = $false
                Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath )"
                Run-Install $packageDirPath
            }
        }

        return $false
    } 

    return $false
}

#endregion


#
# Main script entry point
#

Append-Log INFO "[Start] $ScriptName"

$SilentExit = $false
try {
    $SilentExit = Run
} catch {
    Append-Log ERROR $_.Exception.Message
} finally {
    try {
        if(Test-Path $TempDirPath)
        {
            Append-Log INFO "CleanUp $TempDirPath"
            Remove-Item $TempDirPath -Recurse -Force
        }
    } catch {
        Append-Log ERROR $_.Exception.Message
    }

    Append-Log INFO "[End] $ScriptName"
    if ((-not $SilentExit) -and (-not $Force)) {
        Pause
    }
}
# SIG # Begin signature block
# MIIejwYJKoZIhvcNAQcCoIIegDCCHnwCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUZB5XGFPgs2Fwnld5axnuWhoZ
# SXqgghmZMIIEhDCCA2ygAwIBAgIQQhrylAmEGR9SCkvGJCanSzANBgkqhkiG9w0B
# AQUFADBvMQswCQYDVQQGEwJTRTEUMBIGA1UEChMLQWRkVHJ1c3QgQUIxJjAkBgNV
# BAsTHUFkZFRydXN0IEV4dGVybmFsIFRUUCBOZXR3b3JrMSIwIAYDVQQDExlBZGRU
# cnVzdCBFeHRlcm5hbCBDQSBSb290MB4XDTA1MDYwNzA4MDkxMFoXDTIwMDUzMDEw
# NDgzOFowgZUxCzAJBgNVBAYTAlVTMQswCQYDVQQIEwJVVDEXMBUGA1UEBxMOU2Fs
# dCBMYWtlIENpdHkxHjAcBgNVBAoTFVRoZSBVU0VSVFJVU1QgTmV0d29yazEhMB8G
# A1UECxMYaHR0cDovL3d3dy51c2VydHJ1c3QuY29tMR0wGwYDVQQDExRVVE4tVVNF
# UkZpcnN0LU9iamVjdDCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAM6q
# gT+jo2F4qjEAVZURnicPHxzfOpuCaDDASmEd8S8O+r5596Uj71VRloTN2+O5bj4x
# 2AogZ8f02b+U60cEPgLOKqJdhwQJ9jCdGIqXsqoc/EHSoTbL+z2RuufZcDX65OeQ
# w5ujm9M89RKZd7G3CeBo5hy485RjiGpq/gt2yb70IuRnuasaXnfBhQfdDWy/7gbH
# d2pBnqcP1/vulBe3/IW+pKvEHDHd17bR5PDv3xaPslKT16HUiaEHLr/hARJCHhrh
# 2JU022R5KP+6LhHC5ehbkkj7RwvCbNqtMoNB86XlQXD9ZZBt+vpRxPm9lisZBCzT
# bafc8H9vg2XiaquHhnUCAwEAAaOB9DCB8TAfBgNVHSMEGDAWgBStvZh6NLQm9/rE
# JlTvA73gJMtUGjAdBgNVHQ4EFgQU2u1kdBScFDyr3ZmpvVsoTYs8ydgwDgYDVR0P
# AQH/BAQDAgEGMA8GA1UdEwEB/wQFMAMBAf8wEQYDVR0gBAowCDAGBgRVHSAAMEQG
# A1UdHwQ9MDswOaA3oDWGM2h0dHA6Ly9jcmwudXNlcnRydXN0LmNvbS9BZGRUcnVz
# dEV4dGVybmFsQ0FSb290LmNybDA1BggrBgEFBQcBAQQpMCcwJQYIKwYBBQUHMAGG
# GWh0dHA6Ly9vY3NwLnVzZXJ0cnVzdC5jb20wDQYJKoZIhvcNAQEFBQADggEBAE1C
# L6bBiusHgJBYRoz4GTlmKjxaLG3P1NmHVY15CxKIe0CP1cf4S41VFmOtt1fcOyu9
# 08FPHgOHS0Sb4+JARSbzJkkraoTxVHrUQtr802q7Zn7Knurpu9wHx8OSToM8gUmf
# ktUyCepJLqERcZo20sVOaLbLDhslFq9s3l122B9ysZMmhhfbGN6vRenf+5ivFBjt
# pF72iZRF8FUESt3/J90GSkD2tLzx5A+ZArv9XQ4uKMG+O18aP5cQhLwWPtijnGMd
# ZstcX9o+8w8KCTUi29vAPwD55g1dZ9H9oB4DK9lA977Mh2ZUgKajuPUZYtXSJrGY
# Ju6ay0SnRVqBlRUa9VEwggTmMIIDzqADAgECAhBiXE2QjNVC+6supXM/8VQZMA0G
# CSqGSIb3DQEBBQUAMIGVMQswCQYDVQQGEwJVUzELMAkGA1UECBMCVVQxFzAVBgNV
# BAcTDlNhbHQgTGFrZSBDaXR5MR4wHAYDVQQKExVUaGUgVVNFUlRSVVNUIE5ldHdv
# cmsxITAfBgNVBAsTGGh0dHA6Ly93d3cudXNlcnRydXN0LmNvbTEdMBsGA1UEAxMU
# VVROLVVTRVJGaXJzdC1PYmplY3QwHhcNMTEwNDI3MDAwMDAwWhcNMjAwNTMwMTA0
# ODM4WjB6MQswCQYDVQQGEwJHQjEbMBkGA1UECBMSR3JlYXRlciBNYW5jaGVzdGVy
# MRAwDgYDVQQHEwdTYWxmb3JkMRowGAYDVQQKExFDT01PRE8gQ0EgTGltaXRlZDEg
# MB4GA1UEAxMXQ09NT0RPIFRpbWUgU3RhbXBpbmcgQ0EwggEiMA0GCSqGSIb3DQEB
# AQUAA4IBDwAwggEKAoIBAQCqgvGEqVvYcbXSXSvt9BMgDPmb6dGPdF5u7uspSNjI
# vizrCmFgzL2SjXzddLsKnmhOqnUkcyeuN/MagqVtuMgJRkx+oYPp4gNgpCEQJ0Ca
# WeFtrz6CryFpWW1jzM6x9haaeYOXOh0Mr8l90U7Yw0ahpZiqYM5V1BIR8zsLbMaI
# upUu76BGRTl8rOnjrehXl1/++8IJjf6OmqU/WUb8xy1dhIfwb1gmw/BC/FXeZb5n
# OGOzEbGhJe2pm75I30x3wKoZC7b9So8seVWx/llaWm1VixxD9rFVcimJTUA/vn9J
# AV08m1wI+8ridRUFk50IYv+6Dduq+LW/EDLKcuoIJs0ZAgMBAAGjggFKMIIBRjAf
# BgNVHSMEGDAWgBTa7WR0FJwUPKvdmam9WyhNizzJ2DAdBgNVHQ4EFgQUZCKGtkqJ
# yQQP0ARYkiuzbj0eJ2wwDgYDVR0PAQH/BAQDAgEGMBIGA1UdEwEB/wQIMAYBAf8C
# AQAwEwYDVR0lBAwwCgYIKwYBBQUHAwgwEQYDVR0gBAowCDAGBgRVHSAAMEIGA1Ud
# HwQ7MDkwN6A1oDOGMWh0dHA6Ly9jcmwudXNlcnRydXN0LmNvbS9VVE4tVVNFUkZp
# cnN0LU9iamVjdC5jcmwwdAYIKwYBBQUHAQEEaDBmMD0GCCsGAQUFBzAChjFodHRw
# Oi8vY3J0LnVzZXJ0cnVzdC5jb20vVVROQWRkVHJ1c3RPYmplY3RfQ0EuY3J0MCUG
# CCsGAQUFBzABhhlodHRwOi8vb2NzcC51c2VydHJ1c3QuY29tMA0GCSqGSIb3DQEB
# BQUAA4IBAQARyT3hBeg7ZazJdDEDt9qDOMaSuv3N+Ntjm30ekKSYyNlYaDS18Ash
# U55ZRv1jhd/+R6pw5D9eCJUoXxTx/SKucOS38bC2Vp+xZ7hog16oYNuYOfbcSV4T
# p5BnS+Nu5+vwQ8fQL33/llqnA9abVKAj06XCoI75T9GyBiH+IV0njKCv2bBS7vzI
# 7bec8ckmONalMu1Il5RePeA9NbSwyVivx1j/YnQWkmRB2sqo64sDvcFOrh+RMrjh
# JDt77RRoCYaWKMk7yWwowiVp9UphreAn+FOndRWwUTGw8UH/PlomHmB+4uNqOZrE
# 6u4/5rITP1UDBE0LkHLU6/u8h5BRsjgZMIIE/jCCA+agAwIBAgIQK3PbdGMRTFpb
# MkryMFdySTANBgkqhkiG9w0BAQUFADB6MQswCQYDVQQGEwJHQjEbMBkGA1UECBMS
# R3JlYXRlciBNYW5jaGVzdGVyMRAwDgYDVQQHEwdTYWxmb3JkMRowGAYDVQQKExFD
# T01PRE8gQ0EgTGltaXRlZDEgMB4GA1UEAxMXQ09NT0RPIFRpbWUgU3RhbXBpbmcg
# Q0EwHhcNMTkwNTAyMDAwMDAwWhcNMjAwNTMwMTA0ODM4WjCBgzELMAkGA1UEBhMC
# R0IxGzAZBgNVBAgMEkdyZWF0ZXIgTWFuY2hlc3RlcjEQMA4GA1UEBwwHU2FsZm9y
# ZDEYMBYGA1UECgwPU2VjdGlnbyBMaW1pdGVkMSswKQYDVQQDDCJTZWN0aWdvIFNI
# QS0xIFRpbWUgU3RhbXBpbmcgU2lnbmVyMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8A
# MIIBCgKCAQEAv1I2gjrcdDcNeNV/FlAZZu26GpnRYziaDGayQNungFC/aS42Lwpn
# P0ChSopjNZvQGcx0qhcZkSu1VSAZ+8AaOm3KOZuC8rqVoRrYNMe4iXtwiHBRZmns
# d/7GlHJ6zyWB7TSCmt8IFTcxtG2uHL8Y1Q3P/rXhxPuxR3Hp+u5jkezx7M5ZBBF8
# rgtgU+oq874vAg/QTF0xEy8eaQ+Fm0WWwo0Si2euH69pqwaWgQDfkXyVHOaeGWTf
# dshgRC9J449/YGpFORNEIaW6+5H6QUDtTQK0S3/f4uA9uKrzGthBg49/M+1BBuJ9
# nj9ThI0o2t12xr33jh44zcDLYCQD3npMqwIDAQABo4IBdDCCAXAwHwYDVR0jBBgw
# FoAUZCKGtkqJyQQP0ARYkiuzbj0eJ2wwHQYDVR0OBBYEFK7u2WC6XvUsARL9jo2y
# VXI1Rm/xMA4GA1UdDwEB/wQEAwIGwDAMBgNVHRMBAf8EAjAAMBYGA1UdJQEB/wQM
# MAoGCCsGAQUFBwMIMEAGA1UdIAQ5MDcwNQYMKwYBBAGyMQECAQMIMCUwIwYIKwYB
# BQUHAgEWF2h0dHBzOi8vc2VjdGlnby5jb20vQ1BTMEIGA1UdHwQ7MDkwN6A1oDOG
# MWh0dHA6Ly9jcmwuc2VjdGlnby5jb20vQ09NT0RPVGltZVN0YW1waW5nQ0FfMi5j
# cmwwcgYIKwYBBQUHAQEEZjBkMD0GCCsGAQUFBzAChjFodHRwOi8vY3J0LnNlY3Rp
# Z28uY29tL0NPTU9ET1RpbWVTdGFtcGluZ0NBXzIuY3J0MCMGCCsGAQUFBzABhhdo
# dHRwOi8vb2NzcC5zZWN0aWdvLmNvbTANBgkqhkiG9w0BAQUFAAOCAQEAen+pStKw
# pBwdDZ0tXMauWt2PRR3wnlyQ9l6scP7T2c3kGaQKQ3VgaoOkw5mEIDG61v5MzxP4
# EPdUCX7q3NIuedcHTFS3tcmdsvDyHiQU0JzHyGeqC2K3tPEG5OfkIUsZMpk0uRlh
# dwozkGdswIhKkvWhQwHzrqJvyZW9ljj3g/etfCgf8zjfjiHIcWhTLcuuquIwF4Mi
# KRi14YyJ6274fji7kE+5Xwc0EmuX1eY7kb4AFyFu4m38UnnvgSW6zxPQ+90rzYG2
# V4lO8N3zC0o0yoX/CLmWX+sRE+DhxQOtVxzhXZIGvhvIPD+lIJ9p0GnBxcLJPufF
# cvfqG5bilK+GLjCCBT0wggQloAMCAQICEQDH7PzDLYUk4185wG1anL99MA0GCSqG
# SIb3DQEBCwUAMH0xCzAJBgNVBAYTAkdCMRswGQYDVQQIExJHcmVhdGVyIE1hbmNo
# ZXN0ZXIxEDAOBgNVBAcTB1NhbGZvcmQxGjAYBgNVBAoTEUNPTU9ETyBDQSBMaW1p
# dGVkMSMwIQYDVQQDExpDT01PRE8gUlNBIENvZGUgU2lnbmluZyBDQTAeFw0xNzAz
# MzAwMDAwMDBaFw0xOTExMjUyMzU5NTlaMIGlMQswCQYDVQQGEwJKUDERMA8GA1UE
# EQwIMjMxLTAwMDYxFDASBgNVBAcMC1lva29oYW1hc2hpMQwwCgYDVQQJDAM0MDMx
# IzAhBgNVBAkMGk5ha2FrdSBNaW5hbWluYWthZG9yaSA0IDQzMRIwEAYDVQQKDAlM
# b2lMbyBpbmMxEjAQBgNVBAsMCUxvaUxvIGluYzESMBAGA1UEAwwJTG9pTG8gaW5j
# MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAtG2T2K47U5zZZvxLDYh8
# LHutUZH7oQ26e5cWVBYyF6NPIjVMkyV8heRrtv31apYud2Zahxjuimm1fyHIfHE7
# u06Zi8ZI60rO2okJxBczuHjpnwNrx8QnZTKyhvttRul6TCiQGE9+obOTeZ6/zBUW
# aOSqfGy0B0NNDz2mQGR8/3g8OiT7hrk7i083Ok4um0hXQpKy8zBCFqbUmv882EOk
# z4R4Z6ubP9T0fx/gr+L0771K5sXLhgJkKnq6c5gUs/hHNS+bEU41g669W/ddB9vp
# pD113XaJNh/3vcZeXPhe4DCn0nX/SXMomBYyxAcg25feTVCgdqImoWheLcscSHq1
# LQIDAQABo4IBjTCCAYkwHwYDVR0jBBgwFoAUKZFg/4pN+uv5pmq4z/nmS71JzhIw
# HQYDVR0OBBYEFDaipYiT80rk9VtqBE0x+rGfHZvXMA4GA1UdDwEB/wQEAwIHgDAM
# BgNVHRMBAf8EAjAAMBMGA1UdJQQMMAoGCCsGAQUFBwMDMBEGCWCGSAGG+EIBAQQE
# AwIEEDBGBgNVHSAEPzA9MDsGDCsGAQQBsjEBAgEDAjArMCkGCCsGAQUFBwIBFh1o
# dHRwczovL3NlY3VyZS5jb21vZG8ubmV0L0NQUzBDBgNVHR8EPDA6MDigNqA0hjJo
# dHRwOi8vY3JsLmNvbW9kb2NhLmNvbS9DT01PRE9SU0FDb2RlU2lnbmluZ0NBLmNy
# bDB0BggrBgEFBQcBAQRoMGYwPgYIKwYBBQUHMAKGMmh0dHA6Ly9jcnQuY29tb2Rv
# Y2EuY29tL0NPTU9ET1JTQUNvZGVTaWduaW5nQ0EuY3J0MCQGCCsGAQUFBzABhhho
# dHRwOi8vb2NzcC5jb21vZG9jYS5jb20wDQYJKoZIhvcNAQELBQADggEBAIm22WK0
# aGaovQM8xAgnOeBwhcVun3i8Fkd4XTNkO+jXjarrfhGNkDZDIvJe3Z485dx1JtJC
# pOWAhtOOVozcvHN+qU/YXQ3ZgRFDZE6pftngVK1niwOCdQ52Jj1yvKEpyosSisTz
# D+8u9qy1KbisVAR5K7twCCPAOK2II2reshL1gh1lUq2GJ1/vVaHMG+eXFucqFtT7
# c5j9GTPTKGCpE+ijAga0nYuKtXbazC5IRpXxgRMpmP0XoD3J93VPBbzauhKgCrYv
# YFBijA0iCEvWRNhttpHoOYrd6BMv/2SzTU6BSI3rjLoFqVchPHOgsWcfpr+Q999a
# iNmkA2nrJY30pJwwggXgMIIDyKADAgECAhAufIfMDpNKUv6U/Ry3zTSvMA0GCSqG
# SIb3DQEBDAUAMIGFMQswCQYDVQQGEwJHQjEbMBkGA1UECBMSR3JlYXRlciBNYW5j
# aGVzdGVyMRAwDgYDVQQHEwdTYWxmb3JkMRowGAYDVQQKExFDT01PRE8gQ0EgTGlt
# aXRlZDErMCkGA1UEAxMiQ09NT0RPIFJTQSBDZXJ0aWZpY2F0aW9uIEF1dGhvcml0
# eTAeFw0xMzA1MDkwMDAwMDBaFw0yODA1MDgyMzU5NTlaMH0xCzAJBgNVBAYTAkdC
# MRswGQYDVQQIExJHcmVhdGVyIE1hbmNoZXN0ZXIxEDAOBgNVBAcTB1NhbGZvcmQx
# GjAYBgNVBAoTEUNPTU9ETyBDQSBMaW1pdGVkMSMwIQYDVQQDExpDT01PRE8gUlNB
# IENvZGUgU2lnbmluZyBDQTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEB
# AKaYkGN3kTR/itHd6WcxEevMHv0xHbO5Ylc/k7xb458eJDIRJ2u8UZGnz56eJbNf
# gagYDx0eIDAO+2F7hgmz4/2iaJ0cLJ2/cuPkdaDlNSOOyYruGgxkx9hCoXu1UgNL
# OrCOI0tLY+AilDd71XmQChQYUSzm/sES8Bw/YWEKjKLc9sMwqs0oGHVIwXlaCM27
# jFWM99R2kDozRlBzmFz0hUprD4DdXta9/akvwCX1+XjXjV8QwkRVPJA8MUbLcK4H
# qQrjr8EBb5AaI+JfONvGCF1Hs4NB8C4ANxS5Eqp5klLNhw972GIppH4wvRu1jHK0
# SPLj6CH5XkxieYsCBp9/1QsCAwEAAaOCAVEwggFNMB8GA1UdIwQYMBaAFLuvfgI9
# +qbxPISOre44mOzZMjLUMB0GA1UdDgQWBBQpkWD/ik366/mmarjP+eZLvUnOEjAO
# BgNVHQ8BAf8EBAMCAYYwEgYDVR0TAQH/BAgwBgEB/wIBADATBgNVHSUEDDAKBggr
# BgEFBQcDAzARBgNVHSAECjAIMAYGBFUdIAAwTAYDVR0fBEUwQzBBoD+gPYY7aHR0
# cDovL2NybC5jb21vZG9jYS5jb20vQ09NT0RPUlNBQ2VydGlmaWNhdGlvbkF1dGhv
# cml0eS5jcmwwcQYIKwYBBQUHAQEEZTBjMDsGCCsGAQUFBzAChi9odHRwOi8vY3J0
# LmNvbW9kb2NhLmNvbS9DT01PRE9SU0FBZGRUcnVzdENBLmNydDAkBggrBgEFBQcw
# AYYYaHR0cDovL29jc3AuY29tb2RvY2EuY29tMA0GCSqGSIb3DQEBDAUAA4ICAQAC
# PwI5w+74yjuJ3gxtTbHxTpJPr8I4LATMxWMRqwljr6ui1wI/zG8Zwz3WGgiU/yXY
# qYinKxAa4JuxByIaURw61OHpCb/mJHSvHnsWMW4j71RRLVIC4nUIBUzxt1HhUQDG
# h/Zs7hBEdldq8d9YayGqSdR8N069/7Z1VEAYNldnEc1PAuT+89r8dRfb7Lf3ZQkj
# SR9DV4PqfiB3YchN8rtlTaj3hUUHr3ppJ2WQKUCL33s6UTmMqB9wea1tQiCizwxs
# A4xMzXMHlOdajjoEuqKhfB/LYzoVp9QVG6dSRzKp9L9kR9GqH1NOMjBzwm+3eIKd
# XP9Gu2siHYgL+BuqNKb8jPXdf2WMjDFXMdA27Eehz8uLqO8cGFjFBnfKS5tRr0wI
# SnqP4qNS4o6OzCbkstjlOMKo7caBnDVrqVhhSgqXtEtCtlWdvpnncG1Z+G0qDH8Z
# YF8MmohsMKxSCZAWG/8rndvQIMqJ6ih+Mo4Z33tIMx7XZfiuyfiDFJN2fWTQjs6+
# NX3/cjFNn569HmwvqI8MBlD7jCezdsn05tfDNOKMhyGGYf6/VXThIXcDCmhsu+TJ
# qebPWSXrfOxFDnlmaOgizbjvmIVNlhE8CYrQf7woKBP7aspUjZJczcJlmAaezkhb
# 1LU3k0ZBfAfdz/pD77pnYf99SeC7MH1cgOPmFjlLpzGCBGAwggRcAgEBMIGSMH0x
# CzAJBgNVBAYTAkdCMRswGQYDVQQIExJHcmVhdGVyIE1hbmNoZXN0ZXIxEDAOBgNV
# BAcTB1NhbGZvcmQxGjAYBgNVBAoTEUNPTU9ETyBDQSBMaW1pdGVkMSMwIQYDVQQD
# ExpDT01PRE8gUlNBIENvZGUgU2lnbmluZyBDQQIRAMfs/MMthSTjXznAbVqcv30w
# CQYFKw4DAhoFAKB4MBgGCisGAQQBgjcCAQwxCjAIoAKAAKECgAAwGQYJKoZIhvcN
# AQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUw
# IwYJKoZIhvcNAQkEMRYEFHfutTPzV857oFIncQYh9RP7bjbJMA0GCSqGSIb3DQEB
# AQUABIIBAFHPU9h7KkX8PHfQ0mfjvFR3ErcX+/0HsoiS9OaMsrGbp8oYX9Iblwby
# g6qJIDPyx2gSmjUR10/rhXbXs2K9qqHcMBNMYWaT5zHNPybFTIcxK1JG5fthGrkS
# kDiXT4DrB7TWLoDreVGxmOmZ0/Qlhfadfq8R0vcM0vzSuVj4IHwA+u38+Odcgoh2
# m9nu/imZbxMEf2AU445Ptak6DABB5G2CHm+NPWBiaNXHvoIvDd1JnnD8EvOyTpBs
# FPajf3xJg0DX0m4kc8HmHUHUzRMfnoenAUO3eunB7elp0UTxONBiVbfrywWyc7CQ
# R0/3jqEOb/PG78klV7JSwmAYo1svvNGhggIoMIICJAYJKoZIhvcNAQkGMYICFTCC
# AhECAQEwgY4wejELMAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hl
# c3RlcjEQMA4GA1UEBxMHU2FsZm9yZDEaMBgGA1UEChMRQ09NT0RPIENBIExpbWl0
# ZWQxIDAeBgNVBAMTF0NPTU9ETyBUaW1lIFN0YW1waW5nIENBAhArc9t0YxFMWlsy
# SvIwV3JJMAkGBSsOAwIaBQCgXTAYBgkqhkiG9w0BCQMxCwYJKoZIhvcNAQcBMBwG
# CSqGSIb3DQEJBTEPFw0xOTA3MjIwNzA1MzVaMCMGCSqGSIb3DQEJBDEWBBRcy9AB
# Exe84StnQhh/u83t2OXogDANBgkqhkiG9w0BAQEFAASCAQCFDo3kDsgHJE5vPnlm
# CHswZxbuTIi/5vP/B0UjWyeforqJziTV4NpnAl9fAjbbpiiffgW0jPwls3ZALCAc
# 4ElALVktp0vV49j7UAtvyF2NQXJYulG63LEnGCXLwtMfHGUi9VSy6t5Ds4xNw0hy
# qN4A0ceNUzvlb3blYzgucgEmozTWWhSUJxHF5o7kFUBsKdX0dxbU0eOZ13J18MBE
# u5wd88Q9ZYC4akbIT9kakIPiHkhA7gErdMhjApIKagPFJHFs2dt2IkSHlsFaKpKX
# aJroIVFF/y/wDl31eWRlggByXgA7Fqfm4c7wi/tQU/j1EXaqDrTOr3c2OrO2E37S
# qIUx
# SIG # End signature block
